#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include <linux/videodev2.h>

#include <SDL.h>
#include <SDL_stdinc.h>
#include <SDL_image.h>

#define SCREEN_WIDTH  (640)
#define SCREEN_HEIGHT (480)

#define V4L2_CID_PAN_RELATIVE_LOGITECH  0x0A046D01
#define V4L2_CID_TILT_RELATIVE_LOGITECH 0x0A046D02
#define V4L2_CID_PANTILT_RESET_LOGITECH 0x0A046D03
#define ONE_DEGREE (64)
#define ONE_STEP (ONE_DEGREE * 3)
#define LEFT_RIGHT_MAX (ONE_STEP * 20)
#define UP_DOWN_MAX (ONE_STEP * 10)

static void
die(char *message)
{
  perror(message);
  exit(1);
}

static void
pantilt_reset(int fd)
{
  struct v4l2_ext_control xctrls[1];
  struct v4l2_ext_controls ctrls;
  int ret;

  ctrls.count = 1;
  ctrls.controls = xctrls;
  xctrls[0].id = V4L2_CID_PANTILT_RESET_LOGITECH;
  xctrls[0].value = 3;
  ret = ioctl(fd, VIDIOC_S_EXT_CTRLS, &ctrls);
  if (ret < 0)
    die("pan tilt reset");
  sleep(4);
  return;
}

static Uint32
timer_callback(Uint32 interval, void *arg)
{
  SDL_Event *ev = (SDL_Event *)arg;
  SDL_PushEvent(ev);
  return interval;
}

static void
handle_keyup(SDL_KeyboardEvent *ev,
	     int *done, int fd)
{
  SDL_keysym *keysym;
  SDL_Event quit_ev;
  struct v4l2_ext_control xctrls[2];
  struct v4l2_ext_controls ctrls;
  int ret;

  keysym = &(ev->keysym);
  if (keysym->sym == SDLK_ESCAPE) {
    memset(&quit_ev, 0, sizeof(quit_ev));
    quit_ev.type = SDL_QUIT;
    SDL_PushEvent(&quit_ev);
    return;
  }
  ctrls.count = 2;
  ctrls.controls = xctrls;
  xctrls[0].id = V4L2_CID_PAN_RELATIVE_LOGITECH;
  xctrls[1].id = V4L2_CID_TILT_RELATIVE_LOGITECH;
  switch (keysym->sym) {
  case SDLK_UP:
    xctrls[0].value = 0;
    xctrls[1].value = -ONE_STEP;
    printf("up\n");
    break;
  case SDLK_DOWN:
    xctrls[0].value = 0;
    xctrls[1].value = ONE_STEP;
    printf("down\n");
    break;
  case SDLK_RIGHT:
    xctrls[0].value = -ONE_STEP;
    xctrls[1].value = 0;
    printf("right\n");
    break;
  case SDLK_LEFT:
    xctrls[0].value = ONE_STEP;
    xctrls[1].value = 0;
    printf("left\n");
    break;
  default:
    /* not interested in other keys. return now */
    return;
    break;
  }
  ret = ioctl(fd, VIDIOC_S_EXT_CTRLS, &ctrls);
  if (ret < 0)
    die("VIDIOC_S_EXT_CTRLS");
}

static struct v4l2_ext_control xctrls[] = {
  { .id = V4L2_CID_PAN_RELATIVE_LOGITECH,  },
  { .id = V4L2_CID_TILT_RELATIVE_LOGITECH, },
};
static struct v4l2_ext_controls ctrls = {
  .count    = 2,
  .controls = xctrls,
};

static void
move_cam(int fd)
{
  static int abs_x = 0;
  static int abs_y = 0;
  static int oldx = 0;
  static int oldy = 0;
  int left_right = 0;
  int up_down    = 0;
  int x;
  int y;
  int ret;

  SDL_GetMouseState(&x, &y);

  printf("x: %d, y: %d, ", x, y);

  if ((oldx - x) < 0)
    left_right = -(ONE_STEP);
  else if ((oldx - x) > 0)
    left_right = (ONE_STEP);
  if ((oldy - y) < 0)
    up_down = (ONE_STEP);
  else if ((oldy - y) > 0)
    up_down = -(ONE_STEP);

  printf("abs_x: %d, abs_y: %d, ",
	 abs_x, abs_y);
  printf("left_right: %d, up_down: %d, ",
	 left_right, up_down);
  printf("%s ",
	 left_right < 0 ? "right" :
	 left_right > 0 ? "left" : "-");
  printf("%s\n",
	 up_down < 0 ? "up" :
	 up_down > 0 ? "down" : "-");

  oldx = x;
  oldy = y;

  if (left_right || up_down) {
    if (!((abs_x + left_right) < -LEFT_RIGHT_MAX) &&
	!((abs_x + left_right) > LEFT_RIGHT_MAX) &&
	!((abs_y + up_down) < -UP_DOWN_MAX) &&
	!((abs_y + up_down) > UP_DOWN_MAX)) {

      xctrls[0].value = left_right;
      xctrls[1].value = up_down;
      ret = ioctl(fd, VIDIOC_S_EXT_CTRLS, &ctrls);
      if (ret < 0)
	die("VIDIOC_S_EXT_CTRLS");

      abs_x += left_right;
      abs_y += up_down;
    }
  }
}

static void
handle_userevent(SDL_UserEvent *ev,
		 int fd)
{
  move_cam(fd);
}

int main(int argc, char *argv[])
{
  int fd;
  SDL_Surface *s;
  SDL_Surface *image;
  int ret;
  SDL_Event ev;
  SDL_TimerID id;
  int done = 0;

  ret = SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER);
  if (ret < 0)
    die("sdl init");
  atexit(SDL_Quit);

  memset(&ev, 0, sizeof(ev));
  ev.type = SDL_USEREVENT;
  id = SDL_AddTimer(200, timer_callback, &ev);
  if (!id)
    die("failed to add a timer\n");

  fd = open("/dev/video0", O_RDONLY);
  if (fd < 0)
    die("open");
  pantilt_reset(fd);

  image = IMG_Load("sapporo.jpg");
  if (!image)
    die("failed to load image\n");

  s = SDL_SetVideoMode(SCREEN_WIDTH,
		       SCREEN_HEIGHT,
		       24, SDL_FULLSCREEN);
  if (!s)
    die("failed to set mode\n");

  SDL_BlitSurface(image, NULL, s, NULL);
  SDL_Flip(s);

  while (!done) {
    int ok;
    SDL_Event ev;
    

    ok = SDL_WaitEvent(&ev);
    if (!ok)
      break;
    switch (ev.type) {
    case SDL_KEYUP:
      handle_keyup((SDL_KeyboardEvent *)&ev,
		   &done, fd);
      break;
    case SDL_QUIT:
      printf("quiting...\n");
      done = 1;
      break;
    case SDL_USEREVENT:
      handle_userevent((SDL_UserEvent *)&ev, fd);
      break;
    default:
      break;
    }
  }

  return 0;
}
